/* Routines for UMTS RLC disassembly
 *
 * $Id: packet-rlc.h 30414 2009-10-09 06:37:48Z etxrab $
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

enum rlc_mode {
	RLC_TM,
	RLC_UM,
	RLC_AM
};

#define MAX_RLC_CHANS 64
typedef struct rlc_info
{
	guint32 urnti[MAX_RLC_CHANS];
	guint8 mode[MAX_RLC_CHANS];
	guint8 rbid[MAX_RLC_CHANS];
	gboolean ciphered[MAX_RLC_CHANS];
	gboolean deciphered[MAX_RLC_CHANS];
} rlc_info;
